
*
* WinGen.inc: 	Declarations and Data for arbitrary number of
*		scenarios for the winvest case - include into winmodel.gms.
*
*		This file generates Winvest data for a 3-stage SP.
*               The set "scen" below specifies how many scenarios there are.
*		There are "fan_out" first-stage scenarios - for instance,
*		if "scen" has 250 elements and fan_out = 10, then
*		the scenario tree has 10 branches out of the root node,
*		and 25 brances out of each of the second-stage nodes,
*		for a total of 250 scenarios.
*
* SN, GAMS Summer Schools, 2003.
*
$eolcom #

 option decimals=6, limcol = 100, limrow = 100, solprint = off;

* WinStoch (and others) use the concept of a scenario tree.
* The structure of the scenario tree is determined by the next two sets:
*    Last element of Cscen  indicates the total number of scenarios.
*    Last element of Cnodes indicates the "fan_out" from the root node,
*    as follows: fan_out = card(Cnodes)-card(Cscen)-1 because the 3-stage 
*    scenario tree contains that many nodes at the middle level.
*
*      Example:  Cscen = /scen1 * scen8/ and Cnodes = /node0 * node12/ means:
*    There are 8 scenarios, the root node is node0, nodes node1...node4 are
*    second-stage nodes ("middle level"), and nodes node5...node12 are leaves.
*    fan_out is 4 (4 brances out of the root node).

 set Cscen  /scen1 * scen8/;             alias(s, s1, Cscen);
 set Cnodes /node0 * node12 /;           alias(n, n0, Cnodes);

 set Cmbs  /io2, po7, po70, io90/;  alias(i, Cmbs);
 set Ctime /t0, t1, t2/;                  alias(t, Ctime);


* The following are the parameters (data) of the model, 
* to be generated randomly (more or less based on some underlying model).
* They are generally indexed by (dependent upon) time period and scenario.

  parameter
     yield(i,t,s)    "Instrument yield",
     cash_yield(t,s) "Cash yield",
     liab(t,s)       "Liability incurred",
     val(s)          "Final liability PV";

  parameter
     correl(i)       "Correlation of instrument's returns with cash_yield";

  parameter transcost; transcost = 0.01
  parameter liabs;     liabs = 0;

* For fun, assume the instruments have correlations with cash from -1 to 1:
  correl(i) = -1 + 2 * (ord(i)/card(i));
  display correl;

* Generate these parameters, starting with cash-yield:
     cash_yield(t,s) = max(0.1, normal( 1.05, 1));

* Then the instrument yields, using their correlations to cash_yield:
     yield(i,t,s) =    max(0.1, normal( 1+(cash_yield(t,s)-1) * correl(i), 1));

* Then liab and val (in real life these would be correlated with interest rates):
     liab(t,s) = max(0, normal( 20, 10 ));
     val(s)    = max(0, normal( 60, 40 ));

* Also, don't really need data for t2 (past the horizon):
  cash_yield("t2", s) = 0;
  yield(i, "t2", s)   = 0;
  liab("t2", s)       = 0;

* Introduce the concept of "representative scenario": The parameter
* fanout specifies how many scenarios there are during the first period.
* Each scenario is "represented" by the first scenario passing through
* the same middle node in the tree.
* The set repr(s1,s) allows to easily copy data from representative scenarios
* to all of them.
* 
  parameter fanout, fanout2; 
  fanout  = card(Cnodes)-card(Cscen)-1;
  fanout2 = card(Cscen) / fanout;
  if (fanout*fanout2 <> card(Cscen),
      abort "fanout * fanout2 should equal card(Cscen)";
  )

  set repr(s1, s) "true iff s1 is the representative scenario for s";

* Initialize "repr" using some ugly ord-magic:
  repr(s1, s) = YES $ ((mod(ord(s1), fanout2) = 1) AND
	         ((floor((ord(s1)-1) / fanout2) = floor((ord(s)-1) / fanout2))));
			   
  display repr;

* Finally an extremely important step: Make sure that all data for the
* first period are IDENTICAL for scenarios that are common for that period.
* Now, make sure all scenarios that should have common first-period data get it:
  cash_yield("t0",s) = sum(s1 $ repr(s1,s), cash_yield("t0", s1));
  yield(i, "t0",s)   = sum(s1 $ repr(s1,s), yield(i, "t0", s1));
  liab("t0",s)       = sum(s1 $ repr(s1,s), liab("t0", s1));

  display cash_yield, yield, liab, val;

