
* danmort8.gms: Prototype model for Danish Mortgage Loans.
* This version is an 8-stage stoch. prog., LP.
*
* Apr 20: Added more bonds with better names;
*         Using Rolf's estimates for sigmas and rho:
*           Sigma_s = 0.32, sigma_l = 0.22; rho = 0.33.
$eolcom //
*
* RUN PARAMETERS
$set DISPL  1              // 1 if LOTS OF DISPLAY OUTPUT
$set STAGES 3              // Number of stages, 3 to 8

* The model is a multistage, stochastic optimization model
* for funding mortgage loans in Denmark, using both long
* (20- and 30-year callable bonds) and short (1-year ``Flex") bonds.
* The model is documented in "dan_mort.tex".
*
*       Copyright: Soren S. Nielsen, U. Copenhagen, June 2000.
*
* Structure of this file:
* 1. Scenario tree generation. Nodes in the trinomial tree are numbered
*    from 1 (the root, n1) upwards. Node k's successors are nodes
*    3n-1, 3n, and 3n+1. Code for generating a 3-stage, 4-stage, 5-stage,
*    and 6-stage model is included.
*
* 2. Tree parameter calculation. Using the Nielsen/Ronn interest rate model:
*    "A Two-Factor Model for the Valuation of the T-Bond Futures
*          Contract's Embedded Options", in
*          Advances in Fixed-Income Valuation Modeling and Risk Management,
*          Frank Fabozzi (ed):135--152, 1997.
*
* 3. Yield Curve Calculation. For each node, given the short and long
*    rates at the node, calculate the entire term structure.
*
* 4. Bond Data and Prices: Defines the bond universe and the relevant
*    price calculations: Callable bond prices, annuity calc's for
*    principal and interest payments, etc.
*
* 5. The optimization model. Finding the optimal sequence of portfolios
*    contingent upon the evolution of interest rates. Objective it to
*    minimize expected present value of life-time payments.
*
* 99 Output section, for writing relevant results to a file. Currently
*    not used.


* SECTION 1: Scenario tree generation
*

* Specify the number of levels in the tree, the tau and delta corresponding to
* each level, and calculate the total number of nodes.
*  Tau(lvl):   time corresponding to this level, first one is 0.
*  Delta(lvl): time step between this level and next one

$if NOT "%STAGES%" == "3" $goto STAGS4
* THIS SECTION GENERATES A 3-STAGE PROGRAM WITH 40 NODES
        set node_level /level0 * level3/; alias(node_level, lvl);
        parameter taus(lvl) / level0=0, level1 = 1,
        level2=5, level3 = 30 /;
        set node /n1*n40/; alias(k, anc, node);

$label STAGS4

$if NOT "%STAGES%" == "4" $goto STAGS5
* THIS SECTION GENERATES A 4-STAGE PROGRAM WITH 121 NODES
        set node_level /level0 * level4/; alias(node_level, lvl);
        parameter taus(lvl) / level0=0, level1 = 1,
        level2=5, level3 = 10, level4 = 30 /;
        set node /n1*n121/; alias(k, anc, node);

$label STAGS5

$if NOT "%STAGES%" == "5" $goto STAGS6
* THIS SECTION GENERATES A 5-STAGE PROGRAM WITH 364 NODES
* Solve using cplex or osl
        set node_level /level0 * level5/; alias(node_level, lvl);
        parameter taus(lvl) / level0=0, level1 = 1,
        level2=2, level3 = 5, level4 = 10, level5 = 30 /;
        set node /n1*n364/; alias(k, anc, node);

$label STAGS6

$if NOT "%STAGES%" == "6" $goto STAGS7

* THIS SECTION GENERATES A 6-STAGE PROGRAM WITH 1093 NODES
* Solve using cplex or osl
        set node_level /level0 * level6/; alias(node_level, lvl);
        parameter taus(lvl) / level0=0, level1 = 1, level2 = 2,
        level3=5, level4 = 10, level5 = 20, level6 = 30 /;
        set node /n1*n1093/; alias(k, anc, node);

$label STAGS7

$if NOT "%STAGES%" == "7" $goto STAGS8

* THIS SECTION GENERATES A 7-STAGE PROGRAM WITH 3280 NODES
* Solve using cplex or osl
        set node_level /level0 * level7/; alias(node_level, lvl);
        parameter taus(lvl) / level0=0, level1 = 1, level2 = 2,
        level3 = 3,  level4 = 5, level5 = 10, level6 = 20,
        level7 = 30 /;
        set node /n1*n3280/; alias(k, anc, node);

$label STAGS8

$if NOT "%STAGES%" == "8" $goto STAGS9

* THIS SECTION GENERATES A 8-STAGE PROGRAM WITH 9841 NODES
* Solve using cplex or osl
        set node_level /level0 * level8/; alias(node_level, lvl);
        parameter taus(lvl) / level0=0, level1 = 1, level2 = 2,
        level3 = 3,  level4 = 5, level5 = 10, level6 = 15,
        level7 = 20, level8 = 30 /;
        set node /n1*n9841/; alias(k, anc, node);

$label STAGS9




parameter horizon;   // Time corresponding to leaves
horizon=smax(lvl, taus(lvl)); // (lvl$(ord(lvl) eq card(lvl)), taus(lvl));
parameter deltas(lvl);
deltas(lvl)$(taus(lvl) lt horizon)=taus(lvl+1)-taus(lvl);

parameter num_nodes;
num_nodes=sum(lvl, power(3, ord(lvl)-1));
display horizon, taus, deltas, num_nodes;

* Must manually set the number of nodes to num_nodes: n1 * n(num_nodes).
set root(node) /n1/;

* Check that tree level and number of nodes agree:
abort $ (card(node) ne num_nodes) "Tree level and number of nodes don't agree";

* Set up the ancestor structure among nodes.
* The numbering of nodes is such that node n's successors are nodes 3n-1..3n+1;
* the root node is "n1" with ord("n1")=1.
set ancestor(k, anc);
ancestor(k, anc)=yes $ (3*ord(anc)-1 le ord(k) and 3*ord(anc)+1 ge ord(k));
parameter child(k);  // This node is child number 1, 2 or 3 from its ancestor
child(k)=sum(ancestor(k,anc), ord(k) - 3*ord(anc) + 2);

$if "%DISPL%" == "1"    display ancestor, child;

*
* SECTION 2: Tree parameter Calculation (Nielsen/Ronn model)
*

* Each node has certain data associated with it:
*  The long and short rates at the node,
*  the node
parameters
        long(k), short(k), // the long and short rates at the node
        level(k),          // node's distance from the root
        delta(k),          // time between this node and its successors
        tau(k),            // time corresponding to node
        prob(k),           // probability of this node
        n1(k),             // Number of up-moves leading to this node
        n2(k),             // Number of middle-moves leading to this node
        n3(k);             // Number of down-moves leading to this node

* Basic volatility data:
parameter sigma_r, sigma_l, rho;
sigma_r=0.32; sigma_l = 0.22; rho = 0.33; // BASE CASE FROM ROLF Apr 20 01

* Derived data used in the node calculations:
parameter rho_temp; rho_temp=sqrt(1-sqr(rho));
parameter a1, b1, c1, a2, b2, c2;
a1=sqrt(3/2)*sigma_r; b1 = -a1; c1 = 0;
a2=sigma_l*( sqrt(3/2)*rho + sqrt(1/2)*rho_temp);
b2=sigma_l*(-sqrt(3/2)*rho + sqrt(1/2)*rho_temp);
c2=-sigma_l*sqrt(2)*rho_temp;

* Initialize these for the root, then calculate for the rest of the tree:
parameters short_0, long_0;
short_0=0.044; long_0 = 0.065;

long(root(k))=long_0;
short(root(k))=short_0;
level(root(k))=0;    // Distance from top of tree
delta(root(k))=deltas("level0");
tau(root(k))=0;
prob(root(k))=1;
n1(root(k))=0; n2(root(k)) = 0; n3(root(k)) = 0;

loop {ancestor(k,anc),

        level(k)=level(anc) + 1;
        delta(k)=sum(lvl $ (ord(lvl) = level(k)+1), deltas(lvl));
        tau(k)  =sum(lvl $ (ord(lvl) = level(k)+1), taus(lvl));
        prob(k) =(1/3) * prob(anc);

        n1(k)=n1(anc) + 1 $ (child(k) eq 1);
        n2(k)=n2(anc) + 1 $ (child(k) eq 2);
        n3(k)=n3(anc) + 1 $ (child(k) eq 3);

*       Assume for this simple model that mu_s=mu_l = 0, i.e., no drift:
        short(k)=short_0 * exp((n1(k)*a1+n2(k)*b1+n3(k)*c1)*sqrt(delta(anc)));
        long(k)=long_0  * exp((n1(k)*a2+n2(k)*b2+n3(k)*c2)*sqrt(delta(anc)));

};

* SPD: all the above calculations seem to require we leave
*      the entire node set in up to this point.

*
* SECTION 3: Yield Curves Calculation
*
set years /0*30/; alias(t, years);
parameter yr(t); yr(t)=ord(t)-1; // Used to get the year corresponding to t
parameter curve(k, t); // What the t-year yield at node k is

* Calculate node k's yield curve as a simple linear interpolation on short/long:
curve(k,t)=(yr(t) * long(k) + (30-yr(t)) * short(k)) / 30;


parameter DF(k);   // Node k's discount factor according to root's yield curve
DF(k) = power(1 + sum((root,t) $ (yr(t) eq tau(k)), curve(root,t)), -tau(k));
$if "%DISPL%" == "1" display DF;

*
* SECTION 4: Bond Data and Prices
*
set bonds /Flex1 * Flex5, Fxd10-04 * Fxd10-14,
           Fxd20-04 * Fxd20-14, Fxd30-04 * Fxd30-14/;
alias(i, bonds);
table bond_data(i, *)  // all relevant base data. Times relative to tree start
                issue   matur   period  coupon
Flex1             0       1     30       0.00001
Flex2             0       2     30       0.00001
Flex3             0       3     30       0.00001
Flex4             0       4     30       0.00001
Flex5             0       5     30       0.00001
Fxd10-04          0      10     10      0.04
Fxd10-06          0      10     10      0.06
Fxd10-08          0      10     10      0.08
Fxd10-10          0      10     10      0.10
Fxd10-12          0      10     10      0.12
Fxd10-14          0      10     10      0.14
Fxd20-04          0      20     20      0.04
Fxd20-06          0      20     20      0.06
Fxd20-08          0      20     20      0.08
Fxd20-10          0      20     20      0.10
Fxd20-12          0      20     20      0.12
Fxd20-14          0      20     20      0.14
Fxd30-04          0      30     30      0.04
Fxd30-06          0      30     30      0.06
Fxd30-08          0      30     30      0.08
Fxd30-10          0      30     30      0.10
Fxd30-12          0      30     30      0.12
Fxd30-14          0      30     30      0.14            ;

parameter issue(i); issue(i)=bond_data(i, "issue");
parameter matur(i); matur(i)=bond_data(i, "matur");
parameter coupon(i); coupon(i)=bond_data(i, "coupon");
parameter period(i); period(i)=bond_data(i, "period");
parameter cf(i,t);   // Cash flow of bond i at time t, per 100 face value
cf(i,t)=coupon(i) $ (yr(t) gt issue(i) and yr(t) le matur(i)) +
                         1       $ (yr(t) eq matur(i));

$if "%DISPL%" == "1" display curve, issue, matur, coupon, cf;

* Calc price of each bond at each node.
* NOTE: Check whether problem with correct accounting for coupon payments

parameter price(k,i),   // Price of bond i at node k
call_pr(k,i);           // Price at which issuer can call, min(price, 1.00)
price(k,i)=sum(t $ (yr(t) ge tau(k)),
                      cf(i,t) * (1+curve(k,t)) ** (-(yr(t) - tau(k))));

* Set prices to 0 before they are issued, and after maturity
price(k,i) $ ((tau(k) lt issue(i)) or (tau(k) gt matur(i)))=0;
* Also truncate the price at some upper level:

price(k,i)=min(price(k,i), max(1.07, 1+coupon(i)));
call_pr(k,i)=min(price(k,i), 1.00); // Price at which issuer can call
$if "%DISPL%" == "1" display price, call_pr;


parameter tax_rate; // Tax deduction for interest payments
tax_rate=0.40;

parameter trans_cost;
trans_cost = 0.015;

$if "%DISPL%" == "1" display n1, n2, n3;
display level, prob, long, short;
$if "%DISPL%" == "1" display tau, delta, prob;

$label scenreduction
* now let's shrink the node set
parameter ScenRedParms(*);
parameter ScenRedReport(*);

scalars
* desired SCENRED input parameters
    num_leaves       'number of leaves/scenarios in the original tree',
    num_random       'number of random parameters in a node',
    num_time_steps       'number of times steps',
    num_stages       'number of tree stages',
    red_num_leaves   'desired number of leaves/scenarios in the reduced tree',
    red_percentage   'desired reduction in  reduced tree',
* optional SCENRED input parameters
    reduction_method '0 (default), 1 (forward) or 2 (backward)',
    where_random     '100*inObj+10*inRHS+inMatrix',
    run_time_limit   'upper limit (in sec.) on running time of reduction',
    report_level     '0 (short=default) or 1 (high, additional informations about the tree)';
parameters
    trans_price(i,k)    'Price of bond i at node k',
    trans_call_pr(i,k)  'Price at which issuer can call, min(price, 1.00)';

* define missing SCENRED parameters
num_stages= card(node_level);
num_leaves=sum {k$[tau(k) eq horizon], 1};
red_num_leaves = ceil(num_leaves/10);
num_random=2*card(bonds);

trans_price(i,k) = price(k,i);
trans_call_pr(i,k) = call_pr(k,i);

report_level=  1;

* desired SCENRED input parameters
ScenRedParms('num_leaves') =  num_leaves;
ScenRedParms('num_random') = num_random;
ScenRedParms('num_nodes') = num_nodes;
ScenRedParms('num_time_steps') = num_stages;
* at least one of the following two parameters is desired
ScenRedParms('red_num_leaves') = red_num_leaves;
ScenRedParms('red_percentage') = 0.5;
* optional SCENRED input parameters
ScenRedParms('num_stages') = num_stages;
ScenRedParms('reduction_method') = 1;
ScenRedParms('where_random') = 100;
ScenRedParms('run_time_limit') = 60;
ScenRedParms('report_level') = 0;

execute_unload 'danmortin.gdx',  ScenRedParms,node,ancestor,prob,trans_price, trans_call_pr
file opts / 'scenred.opt' /;
put opts;
put 'log_file =    danmortlog.txt' /;
put 'input_gdx     danmortin.gdx'  /;
put 'output_gdx =  danmortout.gdx' /;
putclose;
execute 'scenred scenred.opt';
execute_load   'danmortout.gdx', ScenRedReport, prob=red_prob,
                              ancestor=red_ancestor;
display scenredreport;
display prob;

$set getreduction
scalar dist;
set sk(node)            'nodes in reduced tree';
set leaves(node)        'Calculate which nodes are leaves (horizon nodes)';

$IF set getreduction $GOTO readreduction

sk(node) = yes;
$GOTO donereduction

$LABEL readreduction

display prob, ScenRedReport;

scalar ugly;
ugly = sum {node$PROB(NODE), 1};
display ugly;
sk(node) = yes$(prob(node));
ugly = card(sk);
display ugly;
ugly = card(ancestor);
display ugly;
ancestor(k,node)$(ancestor(k,node)) = prob(k);
ugly = card(ancestor);
display ugly;

$LABEL donereduction

option clear=leaves;
leaves(sk)=YES $ (tau(sk) eq horizon);

*
* SECTION 5: The optimization model.
*


positive variables          // These variables are all: PRINCIPAL amount
  x(k,i),            // Holding (face value) of bond i out of node k
  p(k,i),            // Purchases (back)
  s(k,i),            // Sales (to raise funds)
  pp(k,i), ip(k,i);  // Interest, principal payment at node k

variables                   // These variables are all: DKK (princip. * prices)
  tot_sales(k),  // Total sales at node k (issue bonds)
  tot_purch(k),  // Total purchases at node k (buy back)
  tot_cash(k);   // Total cash out at node k

variable z;                 // Objective value

* Implement various restrictions:
*  - can't hold a bond outside its life span (issue...maturity)
*  - can't buy at horizon or hold beyond horizon.
s.fx(sk,i) $ (tau(sk) lt issue(i))=0;
p.fx(sk,i) $ (tau(sk) gt matur(i))=0;
x.fx(leaves,i)=0;
s.fx(leaves,i)=0;

* - can't sell bonds that have a price above par:
s.fx(sk,i) $ (price(sk,i) gt 1.00)=0;

equations
  initial(k,i),
  balance(k),
  bal2(k, anc,i),
  principal(k,anc,i),
  interest(k,anc,i),
  objective,
  sales_cnt(k),
  purch_cnt(k),
  cash_cnt(k);

// Individual bond balance - keep track of remaining principal,
// and sales/purchases (which may incur costs).
initial(root(sk),i) ..   s(sk,i)=e= x(sk,i);

bal2(ancestor(sk,anc),i)    ..  // Principal balance constraint
  x(anc,i) - pp(sk,i) + s(sk,i)=e= x(sk,i)$(not leaves(sk)) + p(sk,i);

sales_cnt(sk) ..
        tot_sales(sk)=e= sum(i, price(sk,i) * s(sk,i));
purch_cnt(sk) ..
        tot_purch(sk)=e= sum(i, call_pr(sk,i) * p(sk,i));
cash_cnt(sk) ..
        tot_cash(sk) =e= sum(i, pp(sk,i)  + (1-tax_rate) * ip(sk,i)) +
                tot_purch(sk) - tot_sales(sk) +
                trans_cost * tot_sales(sk) $ (not root(sk));

balance(sk)$(not root(sk)) ..   // no not allow prepayments
        tot_sales(sk)=e= tot_purch(sk);
tot_sales.fx(root)=1000;   // Fix initial loan of 1 mill.
tot_purch.fx(root)=0;

// Annuity calculation here:
//
// PV * (i*(1+i)^n) / ((1+i)^n-1)  is the period payment over
// n periods with period interest i that will pay off a loan of PV.
// The PMT consists of both principal and interest;
// P_j=PMT / ((1+i)^(n-j+1))  is the principal payed at the j'th period,
// j=1,...,n.

// Calculate payment per unit sold, i.e., per x(k,i), per year:
parameter pmt(k,i),  // Total payment per unit principal, node k to succ(k)
        p_pmt(k,i),  // Principal payment period k to successor(k), per unit
        i_pmt(k,i);  // Ditto interest payment


loop{(ancestor(sk,anc), i) $ ((matur(i) gt tau(anc)) and (price(anc,i) gt 0)),
        pmt(anc,i) = coupon(i)*power(1+coupon(i), period(i)-tau(anc)) /
                            (power(1+coupon(i),period(i)-tau(anc)) -1);

        p_pmt(anc,i) = sum(t $ (yr(t) ge issue(i) and yr(t) lt matur(i) and
                                yr(t) ge tau(anc) and yr(t) lt tau(sk)),
                                // Calculate principal payment at year
                                1/power(1+coupon(i), period(i) - yr(t))
                           ) * pmt(anc,i);

        i_pmt(anc,i) = sum(t $ (yr(t) ge issue(i) and yr(t) lt matur(i) and
                                yr(t) ge tau(anc) and yr(t) lt tau(sk)),
                                // Calculate principal payment at year
                                1 - 1/power(1+coupon(i), period(i) - yr(t))
                           ) * pmt(anc,i);
};
$if "%DISPL%" == "1" display pmt, p_pmt, i_pmt;

principal(ancestor(sk,anc),i) .. // principal payment at node k, no prepay
        pp(sk,i) =e=  p_pmt(anc,i) * x(anc,i);
        pp.fx(root,i)=0;

interest(ancestor(sk,anc),i) ..  // interest payment at node k
        ip(sk,i)=e= i_pmt(anc,i) * x(anc,i);
        ip.fx(root,i)=0;

objective ..    // Minimize PV of total expected payments, except at root.
                // Discount by the initial (root) yield curve, DF(k)
        z =e= sum(sk $ (not root(sk)), prob(sk) * tot_cash(sk) * DF(sk));

model Danish /all/;


option iterlim=999999, reslim = 10000, decimals = 6;
option limrow = 0, limcol = 0, solprint = off;

*  option lp=cplex;  // or cplex, for the larger models
solve Danish minimizing z using LP;